@@ -19,7 +19,8 @@
 
 default_ns = globals().copy()
 
-import re,os,sys,string,operator,shutil,getopt,gzip,zlib,stat,traceback
+import re,os,sys,string,operator,shutil,getopt,gzip,zlib,stat,traceback,time
+from glob import glob
 try: # tempfile.mkdtemp is only in python 2.3+
 	from tempfile import mkdtemp
 except ImportError:
@@ -107,6 +108,12 @@ def pathjoin_and_mkdir(*components):
 		os.makedirs(path)
 	return result
 
+def readfile(fn):
+	f = open(fn,'rb')
+	d = f.read()
+	f.close()
+	return d
+
 def writefile_and_reopen(fn,data):
 	"""Write data to file, close, and then reopen readonly, and return the fd.
 
@@ -160,7 +167,6 @@ def test_log_results(cmd,s,o,r,kw):
 	
 
 def expand_cmdline(cmd):
-	from glob import glob
 	argv = []
 	for arg in cmd.split(' '): #bad.  shlex.split would be perfect, but its only in python >=2.3
 		arg = arg.replace('"','') # hack so --foo="bar" works.
@@ -305,6 +311,37 @@ def cfv_stdin_test(cmd,file):
 		r=er
 	test_log_results('stdin/out of '+cmd+' with file '+file,(s1,s2),o1+'\n'+o2,r, None)
 
+def cfv_stdin_progress_test(t,file):
+	s1=s2=None
+	o1=o2=c1=c2=''
+	r=0
+	dir = mkdtemp()
+	try:
+		try:
+			cf1=os.path.join(dir,'cf1.'+t)
+			cf2=os.path.join(dir,'cf2.'+t)
+			s1,o1=runcfv("%s --progress=yes -C -t %s -f %s %s"%(cfvcmd,t,cf1,file))
+			if s1: raise cst_err, 2
+			s2,o2=runcfv("%s --progress=yes -C -t %s -f %s -"%(cfvcmd,t,cf2),stdin=file)
+			if s2: raise cst_err, 3
+			if t!='csv2':#csv2 has only filesize, hence checksum never happens, so no progress
+				x=re.match(re.escape(file)+r' : (\.{20}[-\b.#\\|/]*)[ \r\n]+'+re.escape(cf1)+': (\d+) files, (\d+) OK.  [\d.]+ seconds, [\d.]+K(/s)?$',o1,re.M|re.DOTALL)
+				if not x: raise cst_err, 4
+			x2=re.match(r' : (\.[-\b.#/|\\]*)[\t ]*[ \r\n]+'+re.escape(cf2)+': (\d+) files, (\d+) OK.  [\d.]+ seconds, [\d.]+K(/s)?$',o2,re.M)
+			if not x2: raise cst_err, 5
+			if t=='crc':
+				c1 = readfile(cf1).replace(file,' '*len(file))
+			else:
+				c1 = readfile(cf1).replace(file,'')
+			c2 = readfile(cf2)
+			if c1!=c2: raise cst_err, 6
+		except cst_err, er:
+			r=er
+		test_log_results('progress=yes stdin/out of '+t+' with file '+file,(s1,s2),o1+'\n'+o2+'\n--\n'+c1+'\n'+c2,r, None)
+	finally:
+		shutil.rmtree(dir)
+
+
 def rx_test(pat,str):
 	if re.search(pat,str): return 0
 	return 1
@@ -361,10 +398,18 @@ def cfv_test(s,o, op=op_gt, opval=0):
 		return 0
 	return 1
 
-def cfv_all_test(s,o, files=-2, ok=0, unv=0, notfound=0, badcrc=0, badsize=0, cferror=0, ferror=0, misnamed=0):
+def cfv_status_test(s,o, unv=0, notfound=0, badcrc=0, badsize=0, cferror=0, ferror=0):
 	expected_status = (badcrc and 2) | (badsize and 4) | (notfound and 8) | (ferror and 16) | (unv and 32) | (cferror and 64)
+	if s==expected_status:
+		return 0
+	return 'bad status expected %s got %s'%(expected_status, s)
+
+def cfv_all_test(s,o, files=-2, ok=0, unv=0, notfound=0, badcrc=0, badsize=0, cferror=0, ferror=0, misnamed=0):
+	sresult = cfv_status_test(s,o,unv=unv,notfound=notfound,badcrc=badcrc,badsize=badsize,cferror=cferror,ferror=ferror)
+	if sresult:
+		return sresult
 	x=re.search(rx_StatusLine,tail(o))
-	if s==expected_status and x:
+	if x:
 		if files==-2:
 			files = reduce(operator.add, [ok,badcrc,badsize,notfound,ferror])
 		expected = [files,ok,badcrc,badsize,notfound,ferror,unv,cferror,misnamed]
@@ -372,7 +417,7 @@ def cfv_all_test(s,o, files=-2, ok=0, unv=0, notfound=0, badcrc=0, badsize=0, cf
 		if not filter(icomp, map(None,expected,actual)):
 			return 0
 		return 'expected %s got %s'%(expected,actual)
-	return 'bad status expected %s got %s'%(expected_status, s)
+	return 'status line not found in output'
 
 def cfv_unv_test(s,o,unv=1):
 	x=re.search(rx_Begin+rx_unv+rx_End,tail(o))
@@ -456,6 +501,14 @@ def cfv_version_test(s,o):
 		return 0
 	return 1
 
+def cfv_cftypehelp_test(s,o,expected):
+	if s!=expected:
+		return 1
+	for tname in allfmts()+['auto']:
+		if o.count(tname)<1:
+			return 'type %s not found in output'%tname
+	return 0
+
 def T_test(f, extra=None):
 	cmd=cfvcmd
 	if extra:
@@ -553,6 +606,7 @@ def C_test(f,extra=None,verify=None,t=None,d='data?'):
 	if not t:
 		t=f
 	cfv_stdin_test(cmd+" -t"+f+" -C -f-","data4")
+	cfv_stdin_progress_test(f,'data4')
 	f='test.C.'+f
 	fgz=f+'.gz'
 	try:
@@ -976,23 +1030,46 @@ def test_encoding2():
 	if not BitTorrent:
 		return
 	d = mkdtemp()
+	d2 = mkdtemp()
 	try:
 		cfn = os.path.join(d,u'\u3070\u304B.torrent')
 		shutil.copyfile('testencoding2.torrent.foo', cfn)
-		test_generic(cfvcmd+" -q -T -p "+d, rcurry(status_test,expected=8))
-		test_generic(cfvcmd+" -v -T -p "+d, rcurry(cfv_all_test,ok=0,notfound=4))
+		
+		datafns = [
+			('data1',u'\u2605'),
+			('data2',u'\u2606'),
+			('data3',u'\u262E'),
+			('data4',u'\u2600'),
+		]
+		fnerrs=fnok=0
+		for srcfn,destfn in datafns:
+			try:
+				shutil.copyfile(srcfn, os.path.join(d2,destfn))
+			except (EnvironmentError,UnicodeError):
+				fnerrs=fnerrs+1
+			else:
+				fnok=fnok+1
+		
+		test_generic(cfvcmd+" -q -T -p "+d, rcurry(cfv_status_test,notfound=fnok,ferror=fnerrs))
+		test_generic(cfvcmd+" -v -T -p "+d, rcurry(cfv_all_test,ok=0,notfound=fnok,ferror=fnerrs))
 		bakad = os.path.join(d,u'\u3070\u304B')
 		os.mkdir(bakad)
-		shutil.copyfile('data1',os.path.join(bakad,u'\u2605'))
-		shutil.copyfile('data2',os.path.join(bakad,u'\u2606'))
-		shutil.copyfile('data3',os.path.join(bakad,u'\u262E'))
-		shutil.copyfile('data4',os.path.join(bakad,u'\u2600'))
-		test_generic(cfvcmd+" -q -T -p "+d, rcurry(status_test,expected=0))
-		test_generic(cfvcmd+" -v -T -p "+d, rcurry(cfv_all_test,ok=4))
+		for srcfn,destfn in datafns:
+			try:
+				shutil.copyfile(srcfn, os.path.join(bakad,destfn))
+			except (EnvironmentError,UnicodeError):
+				pass
+		test_generic(cfvcmd+" -q -m -T -p "+d, rcurry(cfv_status_test,ferror=fnerrs))
+		test_generic(cfvcmd+" -v -m -T -p "+d, rcurry(cfv_all_test,ok=fnok,ferror=fnerrs))
+		if not fnerrs:
+			#if some of the files can't be found, checking of remaining files will fail due to missing pieces
+			test_generic(cfvcmd+" -q -T -p "+d, rcurry(cfv_status_test))
+			test_generic(cfvcmd+" -v -T -p "+d, rcurry(cfv_all_test,ok=4))
 	except:
 		import traceback
 		test_log_results('test_encoding2','foobar',''.join(traceback.format_exception(*sys.exc_info())),'foobar',{}) #yuck.  I really should switch this crap all to unittest ...
 	#finally:
+	shutil.rmtree(unicode(d2))
 	shutil.rmtree(unicode(d))
 	
 
@@ -1037,6 +1114,48 @@ def manyfiles_test(t):
 	finally:
 		shutil.rmtree(d)
 
+def specialfile_test(cfpath):
+	if run_internal and ver_fchksum: #current versions of fchksum don't release the GIL, so this deadlocks if doing internal testing and using fchksum.
+		return
+	try:
+		import threading
+	except ImportError:
+		return
+	d = mkdtemp()
+	cfn = os.path.split(cfpath)[1]
+	try:
+		fpath = os.path.join(d,'foo.bar')
+		try:
+			os.mkfifo(fpath)
+		except (AttributeError, EnvironmentError):
+			return
+		shutil.copyfile(cfpath, os.path.join(d, cfn))
+		def pusher(fpath):
+			f=open(fpath,'wb')
+			f.write('a'*0x4000)
+			f.flush()
+			time.sleep(0.1)
+			f.write('b'*0x4000)
+			f.flush()
+			time.sleep(0.1)
+			f.write('c'*0x4000)
+			f.close()
+		t=threading.Thread(target=pusher,args=(fpath,))
+		t.start()
+		s,o=runcfv("%s --progress=yes -T -p %s -f %s"%(cfvcmd,d,cfn))
+		t.join()
+		r=0
+		if s:
+			r=1
+		elif o.count('#')>1:
+			r='count(#) = %s'%(o.count('#'))
+		elif o.count('..'):
+			r=3
+		test_log_results('specialfile_test(%s)'%cfn,s,o,r,None)
+	finally:
+		shutil.rmtree(d)
+	
+
 
 cfvenv=''
 cfvexe=os.path.join(os.pardir,'cfv')
@@ -1227,6 +1346,8 @@ def all_tests():
 		if fmt_cancreate(t) and fmt_available(t):
 			C_funkynames_test(t)
 			manyfiles_test(t)
+	for fn in glob(os.path.join('fifotest','fifo.*')):
+		specialfile_test(fn)
 
 	test_generic(cfvcmd+" -m -v -T -t sfv", lambda s,o: cfv_typerestrict_test(s,o,'sfv'))
 	test_generic(cfvcmd+" -m -v -T -t sfvmd5", lambda s,o: cfv_typerestrict_test(s,o,'sfvmd5'))
@@ -1271,6 +1392,9 @@ def all_tests():
 
 	if run_long_tests:
 		largefile_test()
+	
+	test_generic(cfvcmd+" -t aoeu",rcurry(cfv_cftypehelp_test,1))
+	test_generic(cfvcmd+" -t help",rcurry(cfv_cftypehelp_test,0))
 
 	test_generic(cfvcmd+" -h",cfv_version_test)
 
